# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "1.0.3"

__author__ = "Jamie Haddock"
__email__ = "jhaddock@math.ucla.edu"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2020, Jamie Haddock"

from .ssnmf import SSNMF
